package qsrtdemo.model;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Observable;
import java.util.Observer;
import javax.swing.Timer;
import qsrtdemo.control.ControlPanel;

/**
 * This is the main class of the model. It keeps track of the array being sorted,
 * performs the sorting, and holds a record of all the steps performed and partitions
 * sorted. It also keeps track of which step and partition is currently being
 * looked at.
 *
 * The ControlPanel tells this class when a new array needs to be sorted, and when
 * to change to a different step or partition.
 *
 * This class notifies the view classes when a new array has been sorted, and when
 * the user has changed to a different step or partition.
 *
 * @author Michael Hawes
 */
public class Quicksort extends Observable implements Observer, ActionListener {
    
    public static final Integer UPDATE_BEGIN = new Integer(0);
    public static final Integer UPDATE_STEP_NEXT = new Integer(1);
    public static final Integer UPDATE_STEP_PREV = new Integer(2);
    public static final Integer UPDATE_PARTITION_NEXT = new Integer(3);
    public static final Integer UPDATE_PARTITION_PREV = new Integer(4);
    
    public static final int ORDER_DESCENDING = 0;
    public static final int ORDER_ASCENDING = 1;
    public static final int MAX_PARTITIONS = 100;
    
    private int[] array;
    private int sortOrder;
    private boolean sortDescending;
    private Partition sortTree;
    private Partition[] runOrder;
    private int numParts;
    private int currPartIndex;
    private int currStepIndex;
    private Partition currPart;
    private PartitionStep currStep;
    private PartitionStep endStep;
    private ActionListener endListener;

    private Timer stepTimer = new Timer(STEP_TIME, this);
    private Timer animPadTimer = new Timer(ANIM_PAD_TIME, this);
    public static final int STEP_TIME = 1500;
    public static final int ANIM_PAD_TIME = 500;
    private boolean playing = false;
    private boolean animEnded = true;
    
    public Quicksort() {
        array = new int[1];
    }
    
    public Quicksort(int[] array) {
        sort(array, ORDER_DESCENDING);
    }
    
    public Quicksort(int[] array, int sortOrder) {
        sort(array, sortOrder);
    }
    
    private void sort(int[] array, int sortOrder) {
        this.array = array;
        this.sortOrder = sortOrder;
        if(sortOrder==ORDER_DESCENDING) {
            sortDescending=true;
        } else sortDescending=false;
        
        sortTree = new Partition(array, 0, array.length-1);
        runOrder = new Partition[MAX_PARTITIONS];
        numParts = 0;
        quicksort(sortTree);
        
        if(numParts>0) {
            endStep = getEndStep();
            currStep = runOrder[0].getStep(0);
            currPartIndex = 0;
            currStepIndex = 0;
            currPart = sortTree;
        }
    }
    
    private static int pickPivot(int left, int right) {
        return Math.round((right+left)/2.0f);
    }
    
    private void quicksort(Partition currNode) {
        int left = currNode.getLeftIndex();
        int right = currNode.getRightIndex();
        runOrder[numParts] = currNode;
        numParts++;
        if(left!=right) {
            int pivot = pickPivot(left, right);
            int pivotVal = array[pivot];
            currNode.addStep(new StartPartStep(array, pivot, left, right));
            
            // moving the pivot value to the end of the current selection
            // (does nothing if it's already there)
            if(pivot != right) {
                swap(pivot, right);
                currNode.addStep(new SwapStep(array, right, SwapStep.SWAP_MAKEPIVOTSAFE, pivot, right));
            }
            
            int shouldBeAfter;
            int shouldBeBefore;
            while(true) {
                shouldBeAfter=-1;
                shouldBeBefore=-1;
                // looking for a value that should come after the pivot, starting
                // at the beginning
                for(int i=left; i<right; i++) {
                    if(checkAfterPivot(array[i], pivotVal)) {
                        shouldBeAfter = i;
                        break;
                    }
                }
                // looking for a value that should come before the pivot, starting
                // at the end
                for(int i=right-1; i>=left; i--) {
                    if(checkBeforePivot(array[i], pivotVal)) {
                        shouldBeBefore = i;
                        break;
                    }
                }
                
                if(shouldBeAfter<shouldBeBefore && shouldBeAfter>-1 && shouldBeBefore>-1) {
                    // found values to swap, swap them and continue with
                    // this partition
                    swap(shouldBeAfter, shouldBeBefore);
                    currNode.addStep(new SwapStep(array, right, SwapStep.SWAP_SWAPVALUES, shouldBeAfter, shouldBeBefore));
                } else break;
            }
            
            if(shouldBeAfter<0) {
                // pivot is the last position in this partition, leave it where it is and
                // only a left partition is needed
                currNode.setFinalisedIndex(right);
                currNode.addStep(new ValStep(array, ValStep.VAL_PIVOTEND, right));
                Partition leftNode = new Partition(array, left, right-1);
                currNode.setLeftPart(leftNode);
                quicksort(leftNode);
            } else {
                if(shouldBeBefore<0) {
                    // pivot is the first position in this partition, move it to the start
                    // and only a right partition is needed
                    swap(left, right);
                    currNode.setFinalisedIndex(left);
                    currNode.addStep(new SwapStep(array, left, SwapStep.SWAP_PIVOTSTART, left, right));
                    Partition rightNode = new Partition(array, left+1, right);
                    currNode.setRightPart(rightNode);
                    quicksort(rightNode);
                } else {
                    // partition is done, move pivot to correct position
                    // and partition either side of it
                    swap(shouldBeAfter, right);
                    currNode.setFinalisedIndex(shouldBeAfter);
                    currNode.addStep(new SwapStep(array, shouldBeAfter, SwapStep.SWAP_MAKEPIVOTCORRECT, shouldBeAfter, right));
                    Partition leftNode = new Partition(array, left, shouldBeAfter-1);
                    currNode.setLeftPart(leftNode);
                    quicksort(leftNode);
                    Partition rightNode = new Partition(array, shouldBeAfter+1, right);
                    currNode.setRightPart(rightNode);
                    quicksort(rightNode);
                }
            }
        } else {
            // partition is a single value, no further partitioning needed
            currNode.setFinalisedIndex(left);
            //currNode.addStep(new StartPartStep(array, left, left, right));
            currNode.addStep(new ValStep(array, ValStep.VAL_SINGLEVAL, left));
        }
        currNode.finalise();
    }
    
    private boolean checkBeforePivot(int val, int pivot) {
        if(sortDescending) {
            return val > pivot;
        } else {
            return val < pivot;
        }
    }
    
    private boolean checkAfterPivot(int val, int pivot) {
        if(sortDescending) {
            return val < pivot;
        } else {
            return val > pivot;
        }
    }
    
    private boolean swap(int indexA, int indexB) {
        if(indexA != indexB) {
            int work = array[indexA];
            array[indexA] = array[indexB];
            array[indexB] = work;
            return true;
        } else {
            return false;
        }
    }
    
    private PartitionStep getEndStep() {
        Partition endPart = runOrder[numParts-1];
        return endPart.getStep(endPart.getNumSteps()-1);
    }
    
    private boolean nextStep() {
        boolean success = true;
        currStepIndex++;
        if(currStepIndex>=runOrder[currPartIndex].getNumSteps()) {
            currPartIndex++;
            if(currPartIndex>=numParts) {
                success = false;
                currStepIndex--;
                currPartIndex--;
            } else {
                currPart = runOrder[currPartIndex];
                currStep = currPart.getStep(0);
                currStepIndex = 0;
                if(currStep instanceof SwapStep) animEnded = false;
            }
        } else {
            currStep = runOrder[currPartIndex].getStep(currStepIndex);
            if(currStep instanceof SwapStep) animEnded = false;
        }
        return success;
    }
    
    private boolean prevStep() {
        boolean success = true;
        currStepIndex--;
        if(currStepIndex<0) {
            currPartIndex--;
            if(currPartIndex<0) {
                success = false;
                currStepIndex++;
                currPartIndex++;
            } else {
                currPart = runOrder[currPartIndex];
                currStepIndex = currPart.getNumSteps()-1;
                currStep = currPart.getStep(currStepIndex);
            }
        } else {
            currStep = runOrder[currPartIndex].getStep(currStepIndex);
        }
        return success;
    }
    
    public void update(Observable o, Object arg) {
        playing = false;
        if(arg.equals(ControlPanel.PLAY_STEPS)) {
            if(nextStep()) {
                setChanged();

                if(currStepIndex==0) {
                    notifyObservers(UPDATE_PARTITION_NEXT);
                } else {
                    notifyObservers(UPDATE_STEP_NEXT);
                }

                if(atEndStep() && endListener!=null) {
                    endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                } else {
                    stepTimer.start();
                    playing = true;
                }
            } else if(endListener!=null) {
                endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
            }
        } else if(arg.equals(ControlPanel.STOP_PLAY)) {
            stepTimer.stop();
        } else if(arg.equals(ControlPanel.NEXT_STEP)) {
            if(nextStep()) {
                setChanged();
                if(currStepIndex==0) {
                    notifyObservers(UPDATE_PARTITION_NEXT);
                } else {
                    notifyObservers(UPDATE_STEP_NEXT);
                }
                if(atEndStep() && endListener!=null) {
                    endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                }
            } else if(endListener!=null) {
                endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
            }
        } else if(arg.equals(ControlPanel.PREV_STEP)) {
            if(prevStep()) {
                setChanged();
                if(currStepIndex==(currPart.getNumSteps()-1)) {
                    notifyObservers(UPDATE_PARTITION_PREV);
                } else {
                    notifyObservers(UPDATE_STEP_PREV);
                }
            }
        } else if(arg.equals(ControlPanel.FIRST_STEP)) {
            currStepIndex = 0;
            currPartIndex = 0;
            currPart = runOrder[0];
            currStep = currPart.getStep(0);
            setChanged();
            notifyObservers(UPDATE_PARTITION_NEXT);
        } else if(arg.equals(ControlPanel.FINAL_STEP)) {
            currPartIndex = numParts-1;
            currPart = runOrder[numParts-1];
            currStepIndex = currPart.getNumSteps()-1;
            currStep = currPart.getStep(currStepIndex);
            setChanged();
            notifyObservers(UPDATE_PARTITION_NEXT);
        } else if(arg.equals(ControlPanel.PREV_PART)) {
            if(currPartIndex <= 0) {
                currPartIndex = 0;
            } else {
                currPartIndex--;
            }
            currPart = runOrder[currPartIndex];
            currStepIndex = 0;
            currStep = currPart.getStep(0);
            setChanged();
            notifyObservers(UPDATE_PARTITION_PREV);
        } else if(arg.equals(ControlPanel.NEXT_PART)) {
            if(currPartIndex >= numParts-1) {
                currPartIndex = numParts-1;
                currPart = runOrder[currPartIndex];
                currStepIndex = currPart.getNumSteps()-1;
                currStep = currPart.getStep(currStepIndex);
            } else {
                currPartIndex++;
                currStepIndex = 0;
                currPart = runOrder[currPartIndex];
                currStep = currPart.getStep(0);
            }
            setChanged();
            notifyObservers(UPDATE_PARTITION_NEXT);
        } else {
            Class type = arg.getClass();
            if(type.equals(InitSort.class)) {
                InitSort init = (InitSort)arg;
                sort(init.getArray(), init.getSortOrder());
                setChanged();
                notifyObservers(UPDATE_BEGIN);
            } else if(type.equals(array.getClass())) {
                sort((int[]) arg, ORDER_DESCENDING);
                setChanged();
                notifyObservers(UPDATE_BEGIN);
            }
        }
    }
    
    public Partition getBasePartition() {
        return sortTree;
    }
    
    public PartitionStep getCurrStep() {
        return currStep;
    }
    
    public Partition getCurrPartition() {
        return currPart;
    }
    
    public int getCurrPartNumSteps() {
        return currPart.getNumSteps();
    }
    
    public int getNumParts() {
        return numParts;
    }
    
    public int getCurrStepIndexInPart() {
        return currStepIndex;
    }
    
    public int getCurrPartIndex() {
        return currPartIndex;
    }
    
    public boolean atEndStep() {
        return currStep == endStep;
    }
    
    public int[] getIndicesFinalised() {
        int[] indices = new int[currPartIndex];
        for(int i=0; i<currPartIndex; i++) {
            indices[i] = runOrder[i].getFinalisedIndex();
        }
        return indices;
    }
    
    public int getSortOrder() {
        return sortOrder;
    }
    
    /**
     * Called in order to move the model to a certain partition. The
     * event command should be in the form part[partition index],
     * eg. part0 to move to the first partition.
     * 
     * @param e The event
     */
    public void actionPerformed(ActionEvent e) {
        String cmd = e.getActionCommand();
        Object src = e.getSource();
        if(src.equals(stepTimer)) {
            if(animEnded && !animPadTimer.isRunning()) {
                if(nextStep()) {
                    setChanged();

                    if(currStepIndex==0) {
                        notifyObservers(UPDATE_PARTITION_NEXT);
                    } else {
                        notifyObservers(UPDATE_STEP_NEXT);
                    }

                    if(atEndStep() && endListener!=null) {
                        endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                        playing = false;
                        stepTimer.stop();
                    }
                } else if(endListener!=null) {
                    endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                    playing = false;
                    stepTimer.stop();
                }
            } else stepTimer.stop();
        } else if(src.equals(animPadTimer)) {
            animPadTimer.stop();
            if(playing && !stepTimer.isRunning()) {
                if(nextStep()) {
                    setChanged();

                    if(currStepIndex==0) {
                        notifyObservers(UPDATE_PARTITION_NEXT);
                    } else {
                        notifyObservers(UPDATE_STEP_NEXT);
                    }

                    if(atEndStep() && endListener!=null) {
                        endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                        playing = false;
                    } else stepTimer.start();
                } else if(endListener!=null) {
                    endListener.actionPerformed(new ActionEvent(this, 0, "finished"));
                    playing = false;
                }
            }
        } else if(cmd.equals("animFinished")) {
            animEnded = true;
            if(playing) {
                animPadTimer.start();
            }
        } else if(cmd.startsWith("part")) {
            int partNo = Integer.parseInt(cmd.substring(4));
            if(partNo>-1 && partNo<numParts) {
                currPartIndex = partNo;
                currStepIndex = 0;
                currPart = runOrder[partNo];
                currStep = currPart.getStep(currStepIndex);
                setChanged();
                notifyObservers(UPDATE_PARTITION_NEXT);
            }
        }
    }
    
    /**
     * Called to set an object to inform when the end of the sort is reached
     * 
     * @param l The listener object
     */
    public void setEndListener(ActionListener l) {
        endListener = l;
    }
    
    @Override
    public String toString() {
        return sortTree.toString();
    }
}
