package qsrtdemo.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.font.FontRenderContext;
import java.util.Observable;
import java.util.Observer;
import javax.swing.JPanel;
import javax.swing.Timer;
import qsrtdemo.model.Partition;
import qsrtdemo.model.PartitionStep;
import qsrtdemo.model.Quicksort;
import qsrtdemo.model.StartPartStep;
import qsrtdemo.model.SwapStep;
import qsrtdemo.model.ValStep;

/**
 *
 * @author Michael Hawes
 */
public class StepPanel extends JPanel implements Observer, ActionListener {
    
    private static final BasicStroke stroke = new BasicStroke(4);
    private static final Font font = new Font("Serif", Font.BOLD, 32);
    private static final Font titleFont = new Font("SansSerif", Font.BOLD, 22);
    private static final RenderingHints rh = new RenderingHints(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    private static final String title = "Array at current step";
    private static final float minNodeX = 50;
    private static final float minNodeY = 50;
    private static final float padX = 5;
    private static final float padY = 5;
    private static final float animSpace = 90;
    private static final int titlePad = 30;
    private static final int titleY = 22+(int)padY;
    private static final Color stdColour = Color.black;
    private static final Color biggerColour = new Color(148, 0, 211);
    private static final Color smallerColour = new Color(30, 144, 255);
    private static final Color outsideCurrPartColour = Color.lightGray;
    private static final Color finalPosColour = new Color(152, 251, 152);
    private static final Color pivotColour = Color.blue;
    private static final Color valColour = new Color(0, 220, 0);
    private static final ArrayNodeAnim stdAnim = new StdNodeAnim(stdColour);
    private static final ArrayNodeAnim outsidePartitionAnim = new StdNodeAnim(outsideCurrPartColour);
    private static final ArrayNodeAnim finalPosAnim = new StdNodeAnim(finalPosColour);
    private static final ArrayNodeAnim valAnim = new StdNodeAnim(pivotColour);
    private static final ArrayNodeAnim pivotAnim = new StdNodeAnim(pivotColour);
    private static final ArrayNodeAnim biggerAnim = new StdNodeAnim(biggerColour);
    private static final ArrayNodeAnim smallerAnim = new StdNodeAnim(smallerColour);
    private static ArrayNodeAnim leftAnim;
    private static ArrayNodeAnim rightAnim;
    private static Color leftColour;
    private static Color rightColour;
    private static final int swapSpeed = 300;
    
    private ArrayNode[] nodes;
    private int currPivot;
    private Timer animator;
    private int numNodesAnimating=0;
    private ActionListener animationListener;
    
    /**
     * Method which passes the array to a field
     * requires: an array
     * ensures: the array is set to the corresponding field
     * @param number an integer array
     */
    public StepPanel(int[] number) {
        setArray(number);
        Dimension size = new Dimension();
        size.setSize(minNodeX, minNodeY+padY*2+titlePad+animSpace+stroke.getLineWidth());
        setPreferredSize(size);
    }
    
    public StepPanel(){
        Dimension size = new Dimension();
        size.setSize(minNodeX, minNodeY+padY*2+titlePad+animSpace+stroke.getLineWidth());
        setPreferredSize(size);
    }
    
    /**
     * Shows a visual representation of the array, using Graphics2D
     * requires: an array to be shown
     * ensures: each element of the array is placed in to a Rectangle2D shape,
     *         and is aligned correctly. If the integer is double digits, then
     *         the integer is moved appropriately.
     * @param g graphics object
     */
    public void paintComponent(Graphics g){
        super.paintComponent(g);
        Graphics2D g2 = (Graphics2D)g;
        g2.setRenderingHints(rh);
        g2.setStroke(stroke);
        if(nodes != null) {
            g2.setFont(font);
            
            for(int i=0; i<nodes.length; i++) {
                nodes[i].drawNode(g2);
            }
        }
        g2.setFont(titleFont);
        g2.setColor(Color.black);
        g2.drawString(title, padX, titleY);
    }
    
    public void update(Observable o, Object arg) {
        if(animator!=null) animator.stop();
        Quicksort sort = (Quicksort)o;

        if(sort.getSortOrder()==Quicksort.ORDER_DESCENDING) {
            leftAnim = biggerAnim;
            rightAnim = smallerAnim;
            leftColour = biggerColour;
            rightColour = smallerColour;
        } else {
            rightAnim = smallerAnim;
            leftAnim = biggerAnim;
            leftColour = smallerColour;
            rightColour = biggerColour;
        }

        // Set up the values in the array in their current positions
        int[] array = sort.getCurrStep().getArray();
        setArray(array);

        // Set up appearance and possible animation of the values in the array
        boolean isShowSwap = arg != Quicksort.UPDATE_PARTITION_PREV && arg != Quicksort.UPDATE_STEP_PREV; // Don't show animations if we're going backwards through the steps
        setAnim(sort, isShowSwap);

        repaint();
    }
    
    private void setArray(int[] array) {
        int length = array.length;
        if(length>0) {
            Graphics2D g = (Graphics2D)getGraphics();
            FontRenderContext context = g.getFontRenderContext();
            
            Dimension dimWork;
            float height = minNodeY;
            float width = minNodeX;
            float workHeight;
            float workWidth;
            for(int i=0; i<length; i++) {
                dimWork = ArrayNode.getDimensions(""+array[i], font, context, stroke);
                workHeight = (float)dimWork.getHeight();
                if(workHeight>height) height = workHeight;
                workWidth = (float)dimWork.getWidth();
                if(workWidth>width) width = workWidth;
            }
            ArrayNode.setHeight(height);
            ArrayNode.setWidth(width);
            
            nodes = new ArrayNode[length];
            nodes[0] = new ArrayNode(array[0], padX, padY*2+titlePad, font, context, stroke);
            for(int i=1; i<length; i++) {
                nodes[i] = new ArrayNode(array[i], nodes[i-1], font, context, stroke);
            }
            
            Dimension size = new Dimension();
            size.setSize((width*length)+(ArrayNode.gapX*(length-1))+stroke.getLineWidth()+2*padX, height+stroke.getLineWidth()+animSpace+2*padY+titlePad);
            setPreferredSize(size);
            setMaximumSize(size);
            refreshAlignment();
        } else nodes = null;
    }
    
    private void setAnim(Quicksort sort, boolean isShowSwap) {
        PartitionStep rawStep = sort.getCurrStep();
        Partition currPart = sort.getCurrPartition();

        // Set up appearance of values outside current partition
        for(int i=0; i<currPart.getLeftIndex(); i++) {
            nodes[i].setAnim(outsidePartitionAnim);
        }
        for(int i=currPart.getRightIndex()+1; i<nodes.length; i++) {
            nodes[i].setAnim(outsidePartitionAnim);
        }
        // Set up default appearance of values in current partition
        for(int i=currPart.getLeftIndex(); i<=currPart.getRightIndex(); i++) {
            nodes[i].setAnim(stdAnim);
        }
        // Set up appearance of all values whose positions have already been finalised
        int[] indicesFinalised = sort.getIndicesFinalised();
        for(int i=0; i<indicesFinalised.length; i++) {
            nodes[indicesFinalised[i]].setAnim(finalPosAnim);
        }

        currPivot = rawStep.getPivot();
        Class type = rawStep.getClass();
        if(type.equals(StartPartStep.class)) {
            // For a partition starting step the only non-standard appearance is for the pivot
            nodes[currPivot].setAnim(pivotAnim);

        } else if(type.equals(SwapStep.class)) {
            // For a value swapping step, swapping animations may need to be set up
            // and values may need to be marked as having been found to be lesser
            // than or greater than the pivot

            SwapStep step = (SwapStep)rawStep;
            numNodesAnimating = 2;
            switch(step.getSwapType()) {
                case SwapStep.SWAP_MAKEPIVOTSAFE:
                    // This is the initial swap at the start of most partition executions,
                    // where the pivot is moved to the end of the partition

                    if(isShowSwap) {
                        int safeIndex = step.getIndexB(); // Where the pivot is being moved to
                        int oldIndex = step.getIndexA(); // Where the pivot was to begin with
                        nodes[safeIndex].setAnim(new SwapNodeAnim(pivotColour, nodes[oldIndex], swapSpeed, this));
                        nodes[oldIndex].setAnim(new SwapNodeAnim(stdColour, nodes[safeIndex], swapSpeed, this));
                    } else {
                        nodes[currPivot].setAnim(pivotAnim);
                    }

                    break;
                case SwapStep.SWAP_MAKEPIVOTCORRECT:
                case SwapStep.SWAP_PIVOTSTART:
                    // This is a swap to move the pivot from its position at the
                    // right end of the partition, to its final position in the array.

                    int newPivot = step.getIndexA(); // Where the pivot is being moved to
                    int oldPivot = step.getIndexB(); // Where the pivot was to begin with
                    if(isShowSwap) {
                        nodes[newPivot].setAnim(new SwapNodeAnim(pivotColour, nodes[oldPivot], swapSpeed, this));
                        nodes[oldPivot].setAnim(new SwapNodeAnim(rightColour, nodes[newPivot], swapSpeed, this));
                    } else {
                        nodes[newPivot].setAnim(pivotAnim);
                        nodes[oldPivot].setAnim(rightAnim);
                    }

                    // Mark the rest of the partition's values that we know are
                    // less than or greater than the pivot
                    for(int i=newPivot+1; i<oldPivot; i++) {
                        nodes[i].setAnim(rightAnim);
                    }
                    for(int i=newPivot-1; i>=currPart.getLeftIndex(); i--) {
                        nodes[i].setAnim(leftAnim);
                    }

                    break;
                case SwapStep.SWAP_SWAPVALUES:
                    // This is a swap where we have found a value that needs to be
                    // to the right of the pivot, to the left of a value that needs
                    // to be to the left of the pivot

                    int indexA = step.getIndexA(); // Index of left-side value to be swapped
                    int indexB = step.getIndexB(); // Index of right-side value to be swapped
                    if(isShowSwap) {
                        nodes[indexA].setAnim(new SwapNodeAnim(leftColour, nodes[indexB], swapSpeed, this));
                        nodes[indexB].setAnim(new SwapNodeAnim(rightColour, nodes[indexA], swapSpeed, this));
                    } else {
                        nodes[indexA].setAnim(leftAnim);
                        nodes[indexB].setAnim(rightAnim);
                    }

                    // Mark the rest of the partition's values that we know are
                    // less than or greater than the pivot
                    for(int i=currPart.getLeftIndex(); i<indexA; i++) {
                        nodes[i].setAnim(leftAnim);
                    }
                    for(int i=indexB+1; i<currPart.getRightIndex(); i++) {
                        nodes[i].setAnim(rightAnim);
                    }

                    // Mark the pivot
                    nodes[currPivot].setAnim(pivotAnim);

                    break;
            }

            if(isShowSwap) { // Start swap animation if we're animating
                animator = new Timer(33, this);
                animator.start();
            }

        } else if(type.equals(ValStep.class)) {
            // This is a step where we know all the rest of the values in the
            // partition belong to the right of the pivot, so we're just leaving
            // it in its position at the right end of the partition

            ValStep step = (ValStep)rawStep;
            currPivot = step.getValIndex();
            nodes[currPivot].setAnim(valAnim); // Mark the pivot

            // Mark the rest of the values in the partition
            for(int i=currPivot-1; i>=currPart.getLeftIndex(); i--) {
                nodes[i].setAnim(leftAnim);
            }
        }
    }
    
    public void doLayout() {
        super.doLayout();
        refreshAlignment();
    }
    
    public void refreshAlignment() {
        if(nodes!=null) {
            Dimension dim = getSize();
            float width = (float)dim.getWidth();
            float leftX = nodes[0].getTopLeftX();
            float rightX = nodes[nodes.length-1].getBottomRightX();
            float nodesWidth = rightX - leftX;
            float newXStart = (float)((width-nodesWidth)/2.0);
            float xMove = newXStart - leftX;
            for(int i=0; i<nodes.length; i++) {
                nodes[i].moveX(xMove);
            }
        }
    }
    
    public void actionPerformed(ActionEvent e) {
        Object src = e.getSource();
        if(src.equals(animator)) {
            repaint();
        } else if(src instanceof SwapNodeAnim) {
            numNodesAnimating--;
            if(numNodesAnimating==0) {
                animationListener.actionPerformed(e);
            }
        }
    }

    public void setAnimListener(ActionListener listener) {
        animationListener = listener;
    }
    
}
